/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** adminserv.c - Administrative service functionality
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: adminserv.c,v 1.74 2002/12/09 15:24:38 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "setup.h"
#include <fcntl.h>

#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif

static char errstr[BUFSIZE];

static int a_login(struct User *, int, char **);
static int a_user(struct User *, int, char **);
static int a_shutdown(struct User *, int, char **);
static int a_restart(struct User *, int, char **);
static int a_setpass(struct User *, int, char **);
static int a_quote(struct User *, int, char **);
static int a_cping(struct User *, int, char **);
static int a_cversion(struct User *, int, char **);
static int a_help(struct User *, int, char **);
static int a_rehash(struct User *, int, char **);
static int a_dcc(struct User *, int, char **);
static int a_chat(struct User *, int, char **);
static int a_mem(struct User *, int, char **);

static struct MsgTable admin_table[] =
{
	{ "HELP",	0,0,NULL,	a_help },
	{ "LOGIN",	0,0,NULL,	a_login },
        { "\001DCC",    0,0,NULL,       a_dcc },
        { "CHAT",       0,0,NULL,       a_chat },
	{ "USER",	0,0,NULL,	a_user },
	{ "DIE",	0,0,NULL,	a_shutdown },
	{ "SHUTDOWN",	0,0,NULL,	a_shutdown },
	{ "RESTART",	0,0,NULL,	a_restart },
	{ "SETPASS",	0,0,NULL,	a_setpass },
	{ "QUOTE",	0,0,NULL,	a_quote },
	{ "RAW",	0,0,NULL,	a_quote },
	{ "REHASH",	0,0,NULL,	a_rehash },
	{ "\001PING",	0,0,NULL,	a_cping },
	{ "\001VERSION\001",	0,0,NULL,	a_cversion },
	{ "MEM",	0,0,NULL,	a_mem },
	{ (char *)0,	0,0,NULL,	(int (*)()) 0 }
};

static struct HelpTable help_table[] =
{
        { "LOGIN",      "help/adminserv/login",      0 },
        { "USER",       "help/adminserv/user",       1 },
        { "SHUTDOWN",   "help/adminserv/shutdown",   1 },
        { "RESTART",    "help/adminserv/restart",    1 },
        { "SETPASS",    "help/adminserv/setpass",    1 },
	{ "REHASH",	"help/adminserv/rehash",     1 },
        { "QUOTE",      "help/adminserv/quote",      1 },
        { "RAW",        "help/adminserv/quote",      1 },
	{ "CHAT",	"help/adminserv/chat",       1 },
	{ "MEM",	"help/adminserv/mem",        1 },
        { (char *)0,    (char *)0,                   0 }
};

void adminserv(struct User *u, char *param)
{
	char *parv[MAXPARA];
	int parc = 0;
	int i;
	int status;
	static char buf[BUFSIZE];

        /* Clear the parv */
        for (i = 0; i < MAXPARA; i++)
                parv[i] = NULL;

	/* Clear the error string */
	errstr[0] = '\0';

	strlcpy(buf, param, BUFSIZE);

	if (param != NULL)
		parc = msgtoken(param, parv);

	netinfo.requests++;

	if (!parc)
		return;

	/* We can't show login credentials, so no logging */

        if (IsAOnlyOpers() && !IsOper(u))
	{
		if (parv[0][0] != '\001')
		{
                	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
                       		 "This service is only available to IRCops.");
                	return;
		}
        }

	/* All the preliminary junk is done, now we have to go through 
	** the tables and call the handler for that command
	*/
        for (i=0; admin_table[i].handler; i++)
        {
                if (!strcasecmp(parv[0], admin_table[i].name))
                {
                    status = admin_table[i].handler(u, parc, parv);
                    if (status == 1)
                        log("Error with adminserv command %s: %s", parv[0],
			(errstr) ? errstr : "No error message");
		    /* Matched a value, done, then return */
		    return;
                }
        }
	/* If we get here, we did not get a valid command...
	** Make sure not to say anything to the user if it was a 
	** CTCP that was not matched
	*/
	if (*parv[0] == '\001')
		return;
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Invalid command: %s",
		parv[0]);
	return;
}

/* In all of these functions, parv[0] contains the command, followed
** by the parameters
*/

static int a_help(struct User *u, int parc, char *parv[])
{
	int i;

	spy("%s:  HELP Request by %s for %s", AdminServ.nick, u->nick,
		(parv[1]) ? parv[1] : "NO QUERY");
	sendto_dcc_flag(CF_COMMANDS, "%s:  HELP Request by %s for %s",
		AdminServ.nick, u->nick, (parv[1]) ? parv[1] : "NO QUERY");
	if (parc < 2)
	{
		send_help_file(u->nick, AdminServ.nick, MSG_DEFAULT,
			"help/adminserv/index", "index");
		if (u->myuser != NULL)
			send_help_file(u->nick, AdminServ.nick, MSG_DEFAULT,
				"help/adminserv/aindex", "aindex");
		return 0;
	}

	for (i = 0; help_table[i].filename; i++)
	{
		if (!strcasecmp(parv[1], help_table[i].name))
                {
			if (help_table[i].myuonly)
			{
				if (u->myuser)
					send_help_file(u->nick, AdminServ.nick,
						MSG_DEFAULT,
						help_table[i].filename,
						help_table[i].name);
				else
					privmsg(u->nick, AdminServ.nick,
						MSG_DEFAULT,
						"Permission Denied.");
			}
			else
			{
				send_help_file(u->nick, AdminServ.nick,
					MSG_DEFAULT,
					help_table[i].filename,
					help_table[i].name);
			}

			return 0;
		}
	}

	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"Unknown Help Topic: %s", parv[1]);
	snprintf(errstr, BUFSIZE, "Unknown HELP topic: %s", parv[1]);
	return 1;
}

static int a_login(struct User *u, int parc, char *parv[])
{
        struct MyUser *myu;
#if defined(CRYPT)
        char * crpass;
#endif

        if (parc < 3)
	{
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
			"Syntax: /msg %s LOGIN \37username\37 \37password\37",
			AdminServ.nick);
		strcpy(errstr, "Missing parameters in LOGIN");
                return 1;
        }

        myu = findmyuser(parv[1]);
        if (!myu)
	{
		alert("%s: Failed login attempt (unknown user: %s)",
			AdminServ.nick, parv[1]);
		sendto_dcc_flag(CF_AUTH,
			"%s: Failed login attempt (unknown user: %s)",
			AdminServ.nick, parv[1]);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Unknown User.");
                log("Failed login attempt (unknown user (%s))", parv[1]);
		/* This isn't an error in the code, only to the user */
                return 0;
        }
#if defined(CRYPT)
        crpass = crypt(parv[2], myu->password);
        if (!strcasecmp(myu->username, parv[1]) && 
            !strcmp(myu->password, crpass))
#else
        if (!strcasecmp(myu->username, parv[1]) && 
            !strcmp(myu->password, parv[2]))
#endif /* CRYPT */
	{
                log("%s [%s](%s@%s) sucessfully logged in.", u->nick,
                        myu->username, u->username, u->hostname);
                u->myuser = myu;
		alert("%s: Sucessful login by %s to myunick %s", AdminServ.nick,
			u->nick, myu->username);
		sendto_dcc_flag(CF_AUTH,
			"%s: Sucessful login by %s to myunick %s",
			AdminServ.nick, u->nick, myu->username);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
			"Permission Granted.");
                myu->lastseen = time(NULL);
                myu->flags |= MFLAG_ISON;
                /* We keep the CURRENT person who LOGIN'ed through that
                ** access entry pointed to by the u pointer.  This is used
                ** to obtain information on the user who logged in, ie.
                ** in STATS p
                */
                myu->ptr.u = u;
                return 0;
        }
	else
	{
                log("%s [%s](%s@%s) failed login.", u->nick, myu->username,
			u->username, u->hostname);
		alert("%s: Failed login attempt (invalid password) by %s on "
			"myuser nick %s", AdminServ.nick, u->nick,
			myu->username);
		sendto_dcc_flag(CF_AUTH,
			"%s: Failed login attempt (invalid password) by %s on "
			"myuser nick %s", AdminServ.nick, u->nick,
			myu->username);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Access Denied.");
		/* This isn't a code error either */
		return 0;
        }

	/* NOT REACHED */
	return 0;
}

static int a_user(struct User *u, int parc, char *parv[])
{
        int a = 0;

        if (u->myuser == NULL)
        {
                log("Denied access (USER) to %s [%s].", u->nick, u->hostname);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Access Denied.");
                /* Not a code error */
                return 0;
        }

        if (parc < 2)
	{
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
			"Syntax: /msg %s USER [ADD|DEL|LIST]", AdminServ.nick);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
			"For additional help, /msg %s HELP", AdminServ.nick);
		strcpy(errstr, "Missing parameter in USER");
                return 1;
        }

        if (!strcasecmp(parv[1], "LIST"))
                a = 0;
        else if (!strcasecmp(parv[1], "ADD"))
                a = 1;
        else if (!strcasecmp(parv[1], "DEL"))
                a = 2;
        else
	{
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
			"Unknown Flag: %s", parv[1]);
		/* Not REALLY a code error...but... */
		snprintf(errstr, BUFSIZE, "Unknown flag: %s in USER", parv[1]);
                return 1;
        }

        switch (a)
	{
                case 0:
		{
                        struct MyUser *myu;
			dlink_node *dl;

                        privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
				"UserList:");
			spy("%s: USER LIST Request by %s", AdminServ.nick,
				u->nick);
			sendto_dcc_flag(CF_COMMANDS,
				"%s: USER LIST Request by %s", AdminServ.nick,
				u->nick);
			DLINK_FOREACH(dl, myuhead.head)
			{
				myu = (struct MyUser *) dl->data;
                                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
					"%-20s (%s) -> %s", myu->username,
                                        (IsOperOn(myu)) ? "online" :
					((IsOperDCC(myu)) ? "dcc" : "offline"),
                                        (myu->lastseen == 0) ? "Never Seen" :
						sftime(myu->lastseen));
                        }
                        privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
				"End of List.");
			break;
		}
                case 1:
		{
                        struct MyUser *myu;
			dlink_node *dl;
#if defined(CRYPT)
                        char * salt;
#endif /* CRYPT */
                        if (parc < 4)
			{
                                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
					"Syntax: /msg %s USER ADD "
                                        "<username> <password>",
					AdminServ.nick);
				strcpy(errstr, "Missing parameter in USER");
                                return 1;
                        }
                        myu = findmyuser(parv[2]);
                        if (myu != NULL)
                        {
                                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
                                        "User %s already exists", parv[2]);
                                return 0;
                        }
                        myu = smalloc(sizeof(struct MyUser));
			memcounts.myuser++;
			dl = dlink_create();
                        strlcpy(myu->username, parv[2], MAXNICK + 1);
#if defined(CRYPT)
                        salt = make_salt();
                        strlcpy(myu->password, crypt(parv[3], salt), LINE);
#else
                        strlcpy(myu->password, parv[3], LINE);
#endif /* CRYPT */
                        myu->lastseen = time(NULL);
			dlink_add(myu, dl, &myuhead);
                        privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
				"Added user %s with password %s", parv[2],
				parv[3]);
			alert("%s: USER ADD %s by %s", AdminServ.nick, parv[2],
				u->nick);
			sendto_dcc_flag(CF_AUTH, "%s: USER ADD %s by %s",
				AdminServ.nick, parv[2], u->nick);
                        log("%s [%s](%s@%s) added user %s", u->nick,
				u->myuser->username, u->username, u->hostname,
				myu->username);
                        SaveMyUsers();
			return 0;
			/* NOT REACHED */
			break;
		}
                case 2:
		{
                        if (parc < 3)
			{
                                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
					"Syntax: /msg %s USER DEL <username>",
                                        AdminServ.nick);
				strcpy(errstr, "Missing parameter in USER");
				return 1;
                        }
                        if ((findmyuser(parv[2])) == NULL)
			{
                                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
					"Unable to find user %s", parv[2]);
				alert("%s: USER DEL %s attempt by %s (failed)",
					AdminServ.nick, parv[2], u->nick);
				sendto_dcc_flag(CF_AUTH,
					"%s: USER DEL %s attempt by %s "
					"(failed)", AdminServ.nick, parv[2],
					u->nick);
				snprintf(errstr, BUFSIZE,
					"Unknown myuser: %s", parv[2]);
				return 1;
                        }
                        DeleteMyUser(parv[2]);
                        SaveMyUsers();
                        privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
				"Deleted user %s", parv[2]);
			alert("%s: USER DEL %s by %s", AdminServ.nick, parv[2],
				u->nick);
			sendto_dcc_flag(CF_AUTH, "%s: USER DEL %s by %s",
				AdminServ.nick, parv[2], u->nick);
                        log("%s [%s](%s@%s) deleted user %s", u->nick,
				u->myuser->username, u->username, u->hostname,
				parv[2]);
			return 0;
			/* NOT REACHED */
			break;
		}
        }

	/* NOT REACHED */
	return 0;
}

static int a_shutdown(struct User *u, int parc, char *parv[])
{
	char comm[BUFSIZE];

        if (u->myuser == NULL)
	{
                log("Access Denied (SHUTDOWN) to %s", u->nick);
		alert("%s: SHUTDOWN attempt by %s (failed)", AdminServ.nick,
			u->nick);
		sendto_dcc_flag(CF_NOTICES,
			"%s: SHUTDOWN attempt by %s (failed)", AdminServ.nick,
			u->nick);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Access Denied.");
		/* Not a code error */
                return 0;
        }

	comm[0]='\0';

	ExpandParv(comm, BUFSIZE, 1, parc, parv);

        /* Yes, we are very nasty like this. */
        wallops(settings.name, "%s requested SHUTDOWN: %s", u->nick,
                (comm[0]) ? comm : "No Reason");
	alert("Service Shutdown by %s: %s", u->nick,
		(comm[0]) ? comm : "No Reason");
	sendto_dcc_flag(CF_NOTICES, "Service Shutdown by %s: %s", u->nick,
		(comm[0]) ? comm : "No Reason");
        SendQuit(u, 0, comm);
        sleep(1);
        close(servsock);
	close_dccs();
        SaveStats();
        log("%s [%s](%s) requested SHUTDOWN: %s", u->nick, u->myuser->username,
                u->hostname, (comm) ? comm : "No Reason");
        exit(0);

	/* NOT REACHED */
	return 0;
}

static int a_restart(struct User *u, int parc, char *parv[])
{
	char comm[BUFSIZE];

        if (u->myuser == NULL)
	{
                log("Access Denied (RESTART) to %s", u->nick);
		alert("%s: RESTART Attempt by %s (failed)", AdminServ.nick,
			u->nick);
		sendto_dcc_flag(CF_NOTICES,
			"%s: RESTART Attempt by %s (failed)", AdminServ.nick,
			u->nick);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Access Denied.");
		/* Not a code error */
                return 0;
        }

	comm[0]='\0';

	ExpandParv(comm, BUFSIZE, 1, parc, parv);

        /* Yes, we are very nasty like this. */
        wallops(settings.name, "%s requested RESTART: %s", u->nick,
                (comm[0]) ? comm : "No Reason");
	alert("Service Restart by %s: %s", u->nick,
		(comm[0]) ? comm : "No Reason");
	sendto_dcc_flag(CF_NOTICES, "Service Restart by %s: %s", u->nick,
		(comm[0]) ? comm : "No Reason");
        SendQuit(u, 1, comm);
        sleep(1);
        close(servsock);
	close_dccs();
        SaveStats();
        log("%s [%s](%s) requested RESTART: %s", u->nick, u->myuser->username,
                u->hostname, (comm) ? comm : "No Reason");
        execv(SPATH, NULL);

	/* NOT REACHED */
	return 0;
}

static int a_quote(struct User *u, int parc, char *parv[])
{
#ifdef ENABLE_QUOTE
	char command[BUFSIZE];

        if (u->myuser == NULL)
	{
                log("Access Denied (QUOTE) to %s", u->nick);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Access Denied.");
		/* Not an error in the code */
                return 0;
        }

	ExpandParv(command, BUFSIZE, 1, parc, parv);

        if (command[0] == '\0')
	{
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
			"Please enter a valid command");
		strcpy(errstr, "NULL command attempt in QUOTE");
                return 1;
        }

        sts("%s", command);
        log("%s!%s@%s made me execute %s", u->nick, u->username,
                u->hostname, command);

	return 0;
#else
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "QUOTE/RAW Disabled.");
	return 0;
#endif
}

static int a_setpass(struct User *u, int parc, char *parv[])
{
#if defined(CRYPT)
        char * salt;
#endif /* CRYPT */

        if (u->myuser == NULL)
	{
                log("Access Denied (SETPASS) to %s", u->nick);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Access Denied.");
		/* Not a code error */
                return 0;
        }

        if (parc < 2)
	{
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
			"Please enter a valid password");
		strcpy(errstr, "NULL password attempt in SETPASS");
                return 1;
        }

#if defined(CRYPT)
        salt = make_salt();
        strlcpy(u->myuser->password, crypt(parv[1], salt), LINE);
#else
        strlcpy(u->myuser->password, parv[1], LINE);
#endif /* CRYPT */

        privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Password changed to %s",
		parv[1]);
        /* No I will NOT log the password! */
        log("%s!%s@%s changed his/her password", u->nick, u->username,
                u->hostname);
	SaveMyUsers();

	return 0;
}

static int a_cping(struct User *u, int parc, char *parv[])
{
	char command[BUFSIZE];

	ExpandParv(command, BUFSIZE, 1, parc, parv);

        if (command[0] == '\0')
                return 1;
        sts(":%s NOTICE %s :\001PING %s", AdminServ.nick, u->nick, command);
        log("CTCP PING from %s!%s@%s", u->nick, u->username, u->hostname);

	return 0;
}

static int a_cversion(struct User *u, int parc, char *parv[])
{
        sts(":%s NOTICE %s :\001VERSION %s %s %s -> %s\001",
	     AdminServ.nick, u->nick, stats_version, settings.name,
	     VersionFlags(), stats_date);
        log("CTCP VERSION from %s!%s@%s", u->nick, u->username, u->hostname);

	return 0;
}

static int a_rehash(struct User *u, int parc, char *parv[])
{
        if (u->myuser == NULL)
        {
                log("Access Denied (REHASH) to %s", u->nick);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Access Denied.");
		alert("%s: REHASH Attempt by %s (failed)", AdminServ.nick,
			u->nick);
		sendto_dcc_flag(CF_NOTICES, "%s: REHASH Attempt by %s (failed)",
			AdminServ.nick, u->nick);
                /* Not a code error */
                return 0;
        }

	rehash();

	alert("%s: REHASH by %s", AdminServ.nick, u->nick);
	sendto_dcc_flag(CF_NOTICES, "%s: REHASH by %s", AdminServ.nick,
		u->nick);
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "REHASH complete");

        return 0;
}

static int a_dcc(struct User *u, int parc, char *parv[])
{
	if (parc < 2)
	{
		strcpy(errstr, "Not enough parameters to DCC");
		return 1;
	}

	if (IsDCCOperOnly())
	{
		if (!IsOper(u))
		{
			privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
				"DCC is restricted to opers only.");
			spy("DCC attempt by non-oper %s", u->nick);
			sendto_dcc_flag(CF_AUTH, "DCC attempt by non-oper %s",
				u->nick);
			return 0;
		}
	}

	if (!strcasecmp(parv[1], "CHAT"))
	{
		int sock;
		struct sockaddr_in sa;
		unsigned long hl;
		int optval;
		int flags;

		/* \001DCC CHAT chat longip port\001 */
		if (parc < 5)
		{
			strcpy(errstr, "Not enough parameters to DCC CHAT");
			return 1;
		}

		/* XXX we also manually construct the sock structs */
		if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		{
			strcpy(errstr, "Unable to create socket");
			return 1;
		}

		optval = 1;
		setsockopt(sock, SOL_SOCKET,SO_REUSEADDR, (char *)&optval,
			sizeof(optval));

		/* XXX vhost */
		memset((void *) &sa, '\0', sizeof(sa));
		sa.sin_family = AF_INET;
		sa.sin_port = htons(atoi(parv[4]));
		hl = strtoul(parv[3], NULL, 10);
		sa.sin_addr.s_addr = htonl(hl);

		/* set non blocking, the POSIX way */
		flags = fcntl(sock, F_GETFL, 0);
		flags |= O_NONBLOCK;
		(void) fcntl(sock, F_SETFL, flags);

		connect(sock, (struct sockaddr *) &sa, sizeof(sa));
		add_connection(sock, u->nick, CONN_DCCOUT|CONN_CONNECTING);
		return 0;
	}

	strcpy(errstr, "Unknown DCC option");
	return 1;
}

static int a_chat(struct User *u, int parc, char *parv[])
{
	int sock;
	struct sockaddr_in sa;
	int optval;
	int flags;
	int dcc_port;
	int res = -1;
	char hostname[MAXHOST];
	struct hostent *local_host;
	unsigned long l_ip;

        if (IsDCCOperOnly())
        {
                if (!IsOper(u))
                {
                        privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
                                "DCC is restricted to opers only.");
                        spy("DCC attempt by non-oper %s", u->nick);
                        sendto_dcc_flag(CF_AUTH, "DCC attempt by non-oper %s",
                                u->nick);
                        return 0;
                }
        }

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		strcpy(errstr, "Unable to create socket");
		return 1;
	}

	optval = 1;
	setsockopt(sock, SOL_SOCKET,SO_REUSEADDR, (char *)&optval,
		sizeof(optval));

	for (dcc_port = DCC_LOW_PORT; dcc_port < DCC_HIGH_PORT; dcc_port++)
	{
		memset((void *) &sa, '\0', sizeof(sa));
		sa.sin_family = AF_INET;
		sa.sin_addr.s_addr = INADDR_ANY;
		sa.sin_port = htons(dcc_port);
		if ((res = bind(sock, (struct sockaddr *) &sa,
			sizeof(sa)) >= 0))
		{
			break;
		}
	}

	if (res < 0)
	{
		close(sock);
		privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, 
			"Unable to DCC CHAT");
		strcpy(errstr, "Unable to bind DCC port");
		return 1;
	}

	/* set non blocking, the POSIX way */
	flags = fcntl(sock, F_GETFL, 0);
	flags |= O_NONBLOCK;
	(void) fcntl(sock, F_SETFL, flags);

	if (listen(sock, 5) < 0)
	{
		close(sock);
		privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
                        "Unable to DCC CHAT");
		strcpy(errstr, "Unable to listen()");
		return 1;
	}

	/* accept() will be called when it triggers the readfds in select() */
	add_connection(sock, u->nick, CONN_DCCIN|CONN_CONNECTING);
	/* XXX - vhost here */
	gethostname(hostname, MAXHOST - 1);
	hostname[MAXHOST - 1] = '\0';
	local_host = gethostbyname(hostname);
	(void) memcpy((void *) &l_ip,(void *) local_host->h_addr,
                    sizeof(local_host->h_addr));
	l_ip = htonl(l_ip);

	privmsg(u->nick, AdminServ.nick, MSG_PRIVMSG,
		"\001DCC CHAT chat %lu %d\001", l_ip, dcc_port);

	return 0;
}

static int a_mem(struct User *u, int parc, char *parv[])
{
        if (u->myuser == NULL)
        {
                log("Access Denied (MEM) to %s", u->nick);
                privmsg(u->nick, AdminServ.nick, MSG_DEFAULT, "Access Denied.");
                alert("%s: MEM Attempt by %s (failed)", AdminServ.nick,
                        u->nick);
                sendto_dcc_flag(CF_NOTICES, "%s: MEM Attempt by %s (failed)",
                        AdminServ.nick, u->nick);
                /* Not a code error */
                return 0;
        }

	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"Server count: %d memory: %d bytes", netinfo.servers,
		sizeof(struct Server) * netinfo.servers);
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"User count: %d memory: %d bytes", netinfo.users,
		sizeof(struct User) * netinfo.users);
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"dlink_node count: %d memory: %d bytes", dlink_count,
		dlink_count * sizeof(dlink_node));
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"MyUser count: %d memory: %d bytes", memcounts.myuser,
		memcounts.myuser * sizeof(struct MyUser));
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"Connection count: %d memory: %d bytes", memcounts.connection,
		memcounts.connection * sizeof(struct Connection));
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"Hidden count: %d memory: %d bytes", memcounts.hidden,
		memcounts.hidden * sizeof(struct Hidden));
#ifdef JUPESERV
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"Jupe count: %d memory: %d bytes", memcounts.jupe,
		memcounts.jupe * sizeof(struct Jupe));
#endif
#ifdef SPLITSERV
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"SpServer count: %d memory: %d bytes", memcounts.spserver,
		memcounts.spserver * sizeof(struct SpServer));
#endif
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"Stats count: %d memory: %d bytes", memcounts.stats,
		memcounts.stats * sizeof(struct Stats));
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"TLD count: %d memory: %d bytes", memcounts.tld,
		memcounts.tld * sizeof(struct TLD));
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"TLD->native count: %d memory: %d bytes", memcounts.native,
		memcounts.nativemem);
#ifdef TRACK_GECOS
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"User->gecos count: %d memory: %d bytes", memcounts.gecos,
		memcounts.gecosmem);
#endif
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"Channel count: %d memory: %d bytes", memcounts.channel,
		memcounts.channel * sizeof(struct Channel));
	privmsg(u->nick, AdminServ.nick, MSG_DEFAULT,
		"ChanUser count: %d memory: %d bytes", memcounts.chanuser,
		memcounts.chanuser * sizeof(struct ChanUser));
	return 0;
}
