/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** channels.c - Functions to track channel state
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: channels.c,v 1.27 2002/12/17 14:29:59 wcampbel Exp $
*/

#include <assert.h>
#include "stats.h"
#include "struct.h"
#include "proto.h"

#define MTYPE_NUL	0
#define MTYPE_ADD	1
#define MTYPE_DEL	2

static struct Channel *new_channel(char *);
struct Channel *chanlist[C_TABLE_SIZE];

static void add_channel_to_hash_table(char *, struct Channel *);
static void del_channel_from_hash_table(struct Channel *);

struct CMode
{
	char mode;
	unsigned int value;
};

static struct CMode mode_list[] =
{
	{ 'i',	CMODE_INVITE },
	{ 'm',	CMODE_MOD },
	{ 'n',	CMODE_NOEXT },
	{ 'p',	CMODE_PRIV },
	{ 's',	CMODE_SEC },
	{ 't',	CMODE_TOPIC },
	{ 'a',	CMODE_ANON },
	{ 'q',	CMODE_SRVCHAN },
	{ 'O',	CMODE_OPERONLY },
	{ 'r',	CMODE_REGISTERED },
	{ 'R',	CMODE_REGONLY },
	{ 'j',	CMODE_JUPED },
	{ '\0',	0L }
};

static struct CMode ignore_mode_list[] =
{
	{ 'b',  CMODE_BAN },
	{ 'e',  CMODE_EXEMPT },
	{ 'I',  CMODE_INVEX },
	{ '\0', 0L }
};

static void add_channel_to_hash_table(char *name, struct Channel *c)
{
	c->hash = HASH((unsigned char *) name, C_TABLE_SIZE);
	c->next = chanlist[c->hash];
	chanlist[c->hash] = c;
        debug("add_channel_to_hash_table: %s, %d", name, c->hash);
}

static void del_channel_from_hash_table(struct Channel *c)
{
	struct Channel *tmp, *prev = NULL;

	for (tmp = chanlist[c->hash]; tmp; tmp = tmp->next)
	{
		if (tmp == c)
		{
			if (prev)
				prev->next = tmp->next;
			else
				chanlist[c->hash] = tmp->next;
			tmp->next = NULL;
			return;
		}
		prev = tmp;
	}
}

static struct Channel *new_channel(char *name)
{
	struct Channel *c;

	c = smalloc(sizeof(struct Channel));
	memcounts.channel++;
	if (!name)
		name = "";
	strlcpy(c->name, name, CHANLEN);
	add_channel_to_hash_table(name, c);

	debug("new_channel: %s", name);

	return c;
}

#ifdef IRCD_2_10
struct Channel *AddChannel(char *name)
#else
struct Channel *AddChannel(char *name, long ts)
#endif
{
	struct Channel *c;

	c = findchannel(name);
	if (c != NULL)
	{
		log("Channel %s already exists in the channel table", name);
	}
	else
	{
		c = new_channel(name);
	}
#ifndef IRCD_2_10
	c->ts = ts;
#endif
	
	IncreaseChannels();

	return c;
}

void DelChannel(struct Channel *c)
{

        if (c == NULL)
        {
                log("DelChannel(): %s failed!", (c == NULL) ? "NULL" : c->name);
                return;
        }

	del_channel_from_hash_table(c);
	DecreaseChannels();
	assert(netinfo.channels >= 0);
	assert(c->members.head == NULL);
	assert(c->members.tail == NULL);
	memcounts.channel--;
	free(c);
}

struct Channel *findchannel(char *name)
{
	struct Channel *c;

	debug("findchannel(%s)", name);
	c = chanlist[HASH((unsigned char *) name, C_TABLE_SIZE)];
	while (c && (strcasecmp(c->name, name) != 0))
		c = c->next;
	if (c)
		return c;

	return NULL;
}

void init_channel_hash()
{
	int i;
	struct Channel *c, *b;

	for (i = 0; i < C_TABLE_SIZE; i++)
	{
		c = chanlist[i];
		while (c != NULL)
		{
			b = c->next;
			debug("init_channel_hash() deleting %s", c->name);
			memcounts.channel--;
			free(c);
			c = b;
		}
		chanlist[i] = NULL;
	}
}

void IncreaseChannels()
{
        time_t current = time(NULL);
        debug("Increasing channel count");
        netinfo.channels++;
        if (netinfo.channels > netinfo.maxchannels)
        {
                netinfo.maxchannels = netinfo.channels;
                netinfo.t_maxchannels = current;
        }
        if (netinfo.channels > daily.channels)
        {
                daily.channels = netinfo.channels;
                daily.t_channels = current;
        }
}

void ChannelMode(struct Channel *c, int parc, char *parv[])
{
	int i;
	char *pos = parv[0];
	int matched;
	int parpos = 1;
	int whatt = MTYPE_NUL;
	struct ChanUser *cu;

	if ((pos == NULL) || (*pos == '\0'))
		return;

	/* SJOIN modes of 0 mean no change */
	if (*pos == '0')
		return;

	/* XXX assumed that the parc is correct, bad but the servers
	** do the same
	*/
	for (; *pos != '\0'; pos++)
	{
		matched = FALSE;
		if (*pos == '+')
		{
			whatt = MTYPE_ADD;
			continue;
		}
		if (*pos == '-')
		{
			whatt = MTYPE_DEL;
			continue;
		}
		for (i = 0; mode_list[i].mode != '\0'; i++)
		{
			if (*pos == mode_list[i].mode)
			{
				matched = TRUE;
				if (whatt == MTYPE_ADD)
					c->modes |= mode_list[i].value;
				else
					c->modes &= ~mode_list[i].value;
				break;
			}
		}
		if (matched)
			continue;
		matched = FALSE;
		for (i = 0; ignore_mode_list[i].mode != '\0'; i++)
		{
			if (*pos == ignore_mode_list[i].mode)
			{
				matched = TRUE;
				parpos++;
				break;
			}
		}
		if (matched)
			continue;
		if (*pos == 'l')
		{
			if (whatt == MTYPE_ADD)
			{
				c->modes |= CMODE_LIMIT;
				c->limit = atoi(parv[parpos]);
				parpos++;
			}
			else
			{
				c->modes &= ~CMODE_LIMIT;
				c->limit = 0;
			}
			continue;
		}
		if (*pos == 'k')
		{
			if (whatt == MTYPE_ADD)
			{
				c->modes |= CMODE_KEY;
				strlcpy(c->key, parv[parpos], KEYLEN);
				parpos++;
			}
			else
			{
				c->modes &= ~CMODE_KEY;
				c->key[0] = '\0';
				/* Hybrid typically sends either the key or
				** * on -k, so you should eat a parameter
				*/
				parpos++;
			}
			continue;
		}
		if (*pos == 'o')
		{
			cu = findchanuser(c, finduser(parv[parpos]));
			if (cu == NULL)
			{
				log("CHANNEL ERROR:  MODE %s %co %s", c->name,
					(whatt == MTYPE_ADD) ? '+' : '-',
					parv[parpos]);
				parpos++;
				continue;
			}
			if (whatt == MTYPE_ADD)
				cu->modes |= CMODE_OP;
			else
				cu->modes &= ~CMODE_OP;
			parpos++;
			continue;
		}
		if (*pos == 'v')
		{
			cu = findchanuser(c, finduser(parv[parpos]));
                        if (cu == NULL)
                        {
                                log("CHANNEL ERROR:  MODE %s %cv %s", c->name,
                                        (whatt == MTYPE_ADD) ? '+' : '-',
                                        parv[parpos]);
                                parpos++;
                                continue;
                        }
			if (whatt == MTYPE_ADD)
				cu->modes |= CMODE_VOICE;
			else
				cu->modes &= ~CMODE_VOICE;
			parpos++;
			continue;
		}
		if (*pos == 'h')
		{
			cu = findchanuser(c, finduser(parv[parpos]));
                        if (cu == NULL)
                        {
                                log("CHANNEL ERROR:  MODE %s %ch %s", c->name,
                                        (whatt == MTYPE_ADD) ? '+' : '-',
                                        parv[parpos]);
                                parpos++;
                                continue;
                        }
			if (whatt == MTYPE_ADD)
				cu->modes |= CMODE_HALFOP;
			else
				cu->modes &= ~CMODE_HALFOP;
			parpos++;
			continue;
		}
		debug("ChanMode() mode %c not matched", *pos);
	}
}

char *ModeString(struct Channel *c)
{
	static char modebuf[LINE];
	char limbuf[LIMITLEN];
	char *ptr = modebuf;
	int i;

	*ptr = '+';
	ptr++;

	for (i = 0; mode_list[i].mode != '\0'; i++)
	{
		if (c->modes & mode_list[i].value)
		{
			*ptr = mode_list[i].mode;
			ptr++;
		}
	}
	if (c->modes & CMODE_KEY)
	{
		*ptr = 'k';
		ptr++;
	}
	if (c->modes & CMODE_LIMIT)
	{
		*ptr = 'l';
		ptr++;
	}

	*ptr = '\0';

	if (c->modes & CMODE_KEY)
	{
		strlcat(modebuf, " ", LINE);
		strlcat(modebuf, c->key, LINE);
	}
	if (c->modes & CMODE_LIMIT)
	{
		snprintf(limbuf, LIMITLEN, "%d", c->limit);
		strlcat(modebuf, " ", LINE);
		strlcat(modebuf, limbuf, LINE);
	}

	return modebuf;
}

char *CUModeString(struct ChanUser *cu)
{
	/* increase the size if more modes are used */
	static char modebuf[4];
	char *modeptr = modebuf;

	if (cu->modes & CMODE_OP)
	{
		*modeptr = '@';
		modeptr++;
	}
	if (cu->modes & CMODE_HALFOP)
	{
		*modeptr = '%';
		modeptr++;
	}
	if (cu->modes & CMODE_VOICE)
	{
		*modeptr = '+';
		modeptr++;
	}
	*modeptr = '\0';
	return modebuf;
}

void AddUserToChannel(struct Channel *c, char *userptr)
{
	dlink_node *dl1;
	dlink_node *dl2;
	struct ChanUser *tcu;
	struct ChanUser *cu;
	unsigned int modifiers = 0;
	struct User *u;

	while(*userptr != '\0')
	{
		if (*userptr == '@')
		{
			modifiers |= CMODE_OP;
			userptr++;
			continue;
		}
		if (*userptr == '+')
		{
			modifiers |= CMODE_VOICE;
			userptr++;
			continue;
		}
		if (*userptr == '%')
		{
			modifiers |= CMODE_HALFOP;
			userptr++;
			continue;
		}
		/* If I get here, I've parsed everything and it SHOULD be
		** the nick.
		*/
		break;
	}

	u = finduser(userptr);

	/* XXX */
	assert(u != NULL);

	tcu = findchanuser(c, u);
	if (tcu != NULL)
	{
		log("AddUserToChannel: %s already in %s", userptr, c->name);
		/* Since this could be from an OPME, still set the modifiers */
		tcu->modes |= modifiers;
		return;
	}

	dl1 = dlink_create();
	dl2 = dlink_create();
	cu = smalloc(sizeof(struct ChanUser));
	memcounts.chanuser++;

	cu->chan = c;
	cu->user = u;
	cu->modes |= modifiers;

	dlink_add(cu, dl1, &c->members);
	dlink_add(cu, dl2, &u->channels);
	c->nummembers++;
}

void AddUserFlagToChannel(struct Channel *c, struct User *u, unsigned int flags)
{
        dlink_node *dl1;
        dlink_node *dl2;
	struct ChanUser *tcu;
        struct ChanUser *cu;

	tcu = findchanuser(c, u);
	if (tcu != NULL)
	{
		log("AddUserFlagToChannel: %s already in %s", u->nick, c->name);
		/* Since this could be from an OPME, still set the modifiers */
		tcu->modes |= flags;
		return;
	}

	dl1 = dlink_create();
	dl2 = dlink_create();
	cu = smalloc(sizeof(struct ChanUser));
        memcounts.chanuser++;

        cu->chan = c;
        cu->user = u;
        cu->modes |= flags;

        dlink_add(cu, dl1, &c->members);
        dlink_add(cu, dl2, &u->channels);
        c->nummembers++;
}

/* Side effect, free's the ChanUser struct */
void DelUserFromChannel(struct Channel *c, struct User *u)
{
	dlink_node *dl;
	struct ChanUser *cu;

	DLINK_FOREACH(dl, c->members.head)
	{
		cu = (struct ChanUser *) dl->data;
		if (cu->user == u)
		{
			dlink_delete(dl, &c->members);
			free(cu);
			memcounts.chanuser--;
			dlink_free(dl);
			c->nummembers--;
			if (c->nummembers == 0)
				DelChannel(c);
			return;
		}
	}
	log("DelUserFromChannel(%s, %s) failed", c->name, u->nick);
}

void UnlinkChannelFromUser(struct Channel *c, struct User *u)
{
	dlink_node *dl;
	struct ChanUser *cu;

	DLINK_FOREACH(dl, u->channels.head)
	{
		cu = (struct ChanUser *) dl->data;
		if (cu->chan == c)
		{
			dlink_delete(dl, &u->channels);
			dlink_free(dl);
			return;
		}
	}
	log("UnlinkUserFromChannel(%s, %s) failed", c->name, u->nick);
}

struct ChanUser *findchanuser(struct Channel *c, struct User *u)
{
	dlink_node *dl;
	struct ChanUser *cu;

	if ((c == NULL) || (u == NULL))
		return NULL;

	DLINK_FOREACH(dl, c->members.head)
	{
		cu = (struct ChanUser *) dl->data;
		if (cu->user == u)
			return cu;
	}

	return NULL;
}
