/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** main.c - main() and ancilliary functions
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: main.c,v 1.112 2002/12/19 19:45:08 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include <signal.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <locale.h>
#include "event.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

int servsock = -1;
int need_rehash;

static void start();
static void CmdLine(int, char **);
static void PrintVersion();
static void PrintHelp();
static void setup_corefile(void);
static void InitMemCounts(void);

int main(int argc, char *argv[])
{
	int forked = 0;
	FILE *fp;
	time_t current = time(NULL);

	/* Set up localization of times and dates */
	(void) setlocale(LC_TIME, "");

	/* LC_CTYPE must remain in the "C" locale or else toupper() and
	** tolower() will be localized.  In the "C" locale, these functions
	** provide an ASCII casemap.
	**
	** The LC_* variables default to being in the "C" locale when
	** a program starts.  Therefore, we do not need an explicit
	** setlocale(LC_CTYPE, "C");
	*/
	
	/* Some systems default to not dumping core...
	** we want the cores...
	*/
	setup_corefile();

	me.cmdflags = 0;
	need_rehash = 0;

	CmdLine(argc, argv);

	log("Statistics Started (%s %s).", stats_version, VersionFlags());
	me.t_start = current;

	/* chdir returns 0 on sucess, -1 on failure */
	if (chdir(DPATH) != 0)
	{
		printf("Unable to chdir to DPATH, did you edit config.h?\n");
		printf("Fatal Error, exiting\n");
		exit(1);
	}

	InitDefaults();
	InitLists();
	InitMemCounts();
	ConfLoad();
	VerifyConf();
	InitMessageStrings();
	InitMsgHash();
	InitHTMLArray();
#ifndef IRCD_2_10
	me.capabs = generate_capab_string();
#endif

	/* This should ONLY be run once */
	srandom(current); 

	/* Zero out the byte counters */
	netinfo.inb = 0;
	netinfo.outb = 0;
#ifdef DIE_ON_MASS_COLLIDE
	netinfo.collides = 0;
#endif

	LoadStats();
	LoadTLD();
	LoadMyUsers();

	eventInit();

	if (!IsDebug() && !IsNoFork())
		forked=fork();
	if (forked)
	{
		fp = fopen(PIDFILE, "w");
		fprintf(fp, "%i", forked);
		fclose(fp);
		return 0;
	}

	if (!IsDebug() && !IsNoFork())
	{
		freopen("/dev/null", "r", stdin);
		freopen("/dev/null", "w", stdout);
		freopen("stderr.log", "a", stderr);
	}

#ifdef SPLITSERV
	SpInitTables();
#endif

#if defined(SPLITSERV)
	is_delayed = 1;
#endif
#if defined(SPYMODE) || defined(ALERTMODE)
	ad_delayed = 1;
#endif

        /* Setup signals */
        signal(SIGTERM, SigTerm);
	signal(SIGHUP, SigHup);
	signal(SIGUSR1, SigUsr1);
	signal(SIGINT, SigInt);

	/* On FreeBSD at least, gdb will cause a process to core
	** when you detach the debugger from a running Stats.
	** This prevents that from happening.  Reasoning?
	*/
	signal(SIGTRAP, SIG_IGN);

	/* I've been told that SIGPIPE can occur on sockets, just for
	** safety, I'll ignore them.
	*/
	signal(SIGPIPE, SIG_IGN);

	setsid();

	while(TRUE)
	{
		start();
	}

	/* NOT REACHED */
	return 1;
}

static void start()
{
	netinfo.users = netinfo.opers = netinfo.servers = netinfo.channels = 0;
	init_server_hash();
	init_user_hash();
	init_channel_hash();
	init_tld();

	log("Connecting to %s:%d", settings.uplink, settings.port);
	if (servsock > 0)
	{
		close(servsock);
		delete_connection(servsock);
	}

	ConnectTo(settings.uplink, settings.port);

	StartInitialEvents();
	read_loop();
}

void ServTermination(void)
{
	FlushEvents();

	log("Reconnecting to the server in %d seconds", settings.r_time);
	/* Here I need to flush the server/user table */
	if (me.s != NULL)
	{
		FlushServer(me.s->name);
		me.s = NULL;
	}
	else
	{
		debug("Invalid me.s");
	}
	/* To prevent multiple operlist entries, delete the services
	** users so they are recreated fresh
	*/
	DelUser(AdminServ.nick);
#ifdef SPLITSERV
	DelUser(SplitServ.nick);
#endif
#ifdef STATSERV
	DelUser(StatServ.nick);
#endif
#ifdef JUPESERV
	DelUser(JupeServ.nick);
#endif

	eventAdd("Server Connect", evh_ConnectServer, NULL, settings.r_time);
#if defined(SPLITSERV)
        /* Reset the variables used in initialization to prevent the
        ** service from spewing large chunks of data out to the channel
        */
        is_delayed = 1;
#endif
#if defined(SPYMODE) || defined(ALERTMODE)
	ad_delayed = 1;
#endif
}

static void CmdLine(int argc, char *argv[])
{
	char c;

	while( (c=getopt(argc, argv, "dvhn")) != -1)
	{
		switch (c)
		{
			case 'd':
				me.cmdflags |= CMD_DEBUG;
				break;
			case 'v':
				PrintVersion();
				exit(0);
				/* NOT REACHED */
				break;
			case 'h':
				PrintHelp();
				exit(0);
				/* NOT REACHED */
				break;
			case 'n':
				me.cmdflags |= CMD_NOFORK;
				break;
		}
	}
}

static void PrintVersion()
{
	printf("Stats version %s\nSPATH: %s\n", stats_version,
		SPATH);
	printf("Stats NOT Started.\n");
}

static void PrintHelp()
{
	printf("stats [-n|-d|-h|-v]\n-n Do not fork into the background\n");
	printf("-d Set debugging mode\n-h Print this help screen\n");
	printf("-v Print the version and exit\n");
	printf("Stats NOT Started.\n");
}

/*
 * setup_corefile
 *
 * inputs       - nothing
 * output       - nothing
 * side effects - setups corefile to system limits.
 * -kre
 *
 * Stolen from Hyb6.2 - Hwy
 */
static void setup_corefile(void)
{
  struct rlimit rlim; /* resource limits */

  /* Set corefilesize to maximum */
  if (!getrlimit(RLIMIT_CORE, &rlim))
  {
    rlim.rlim_cur = rlim.rlim_max;
    setrlimit(RLIMIT_CORE, &rlim);
  }
}

static void InitMemCounts(void)
{
	memcounts.myuser = 0;
	memcounts.connection = 0;
	memcounts.hidden = 0;
#ifdef JUPESERV
	memcounts.jupe = 0;
#endif
#ifdef SPLITSERV
	memcounts.spserver = 0;
#endif
	memcounts.stats = 0;
	memcounts.tld = 0;
	memcounts.native = 0;
	memcounts.nativemem = 0;
	memcounts.channel = 0;
	memcounts.chanuser = 0;
}
