/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** pubstats.c - Handle public (channel) StatServ commands
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: pubstats.c,v 1.6 2002/11/23 23:05:03 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "match.h"
#include "text.h"

#ifdef PUBSTATS

static char errstr[BUFSIZE];

static int p_map(struct User *, int, char **);
static int p_version_map(struct User *, int, char **);
static int p_tld(struct User *, int, char **);
static int p_tld_map(struct User *, int, char **);
static int p_country(struct User *, int, char **);
static int p_daily(struct User *, int, char **);
static int p_netstats(struct User *, int, char **);
static int p_url(struct User *, int, char **);
static int p_help(struct User *, int, char **);

static struct MsgTable pub_table[] =
{
	{ PUB_MAP,	0,0,NULL,	p_map },
	{ PUB_VERSIONMAP,0,0,NULL,	p_version_map },
	{ PUB_TLD,	0,0,NULL,	p_tld },
	{ PUB_TLDMAP,	0,0,NULL,	p_tld_map },
	{ PUB_COUNTRY,	0,0,NULL,	p_country },
	{ PUB_DAILY,	0,0,NULL,	p_daily },
	{ PUB_NETSTATS,	0,0,NULL,	p_netstats },
	{ PUB_URL,	0,0,NULL,	p_url },
	{ PUB_HELP,	0,0,NULL,	p_help },
	{ NULL,		0,0,NULL,	NULL }
};

void pubstats(struct User *u, char *target, char *param)
{
	char *parv[MAXPARA];
	int parc = 0;
	int i;
	int status;

	/* Clear the parv */
	for (i = 0; i < MAXPARA; i++)
		parv[i] = NULL;

	/* Clear the error string */
	errstr[0] = '\0';

	if (param != NULL)
		parc = msgtoken(param, parv);

	netinfo.requests++;

	if (parc == 0)
		return;

	for (i=0; pub_table[i].handler; i++)
	{
		if (!strcasecmp(parv[0], pub_table[i].name))
		{
			status = pub_table[i].handler(u, parc, parv);
			if (status == 1)
				log("Error with statserv pub command %s: %s",
					parv[0],
					(errstr) ? errstr : "No error message");
			/* Matched a value, done, then return */
			return;
		}
	}

	return;
}

/* All of these commands are subsets of the full StatServ commands, optimized
** for channel use, showing no hidden servers or IPs.
*/

/* In all of these functions, parv[0] contains the command, followed
** by the parameters
*/

static int p_map(struct User *u, int parc, char *parv[])
{
	struct Server *s;
	int i;

	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"Network Map [NAME] [USERS/MAX] "
		"[OPERS/MAX] [PERCENT]");
	for (i = 0; i < S_TABLE_SIZE; i++)
	{
		for (s = serverlist[i]; s; s = s->next)
		{
			if (!IsSHide(s))
			{
				if (IsPubDelayed()) select_delay();
#ifdef LARGE_NET
				privmsg(settings.pub_chan, StatServ.nick,
					MSG_PRIVMSG,
					"%-30s [%3d/%3d] [%2d/%2d] [%2.0f%%]",
#else
				privmsg(settings.pub_chan, StatServ.nick,
					MSG_PRIVMSG,
					"%-30s [%2d/%2d] [%2d/%2d] [%2.0f%%]",
#endif
					s->name, s->users, s->stats->maxusers,
					s->opers, s->stats->maxopers,
					(float)s->users / (float)netinfo.users
						* 100);
			}
		}
	}
	if (IsPubDelayed()) select_delay();
#ifdef LARGE_NET
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"%-30s [%4d/%4d] [%3d/%3d] [100%%]",
#else
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"%-30s [%2d/%2d] [%2d/%2d] [100%%]",
#endif
		"Totals:", netinfo.users, netinfo.maxusers, netinfo.opers,
		netinfo.maxopers);

	return 0;
}

static int p_version_map(struct User *u, int parc, char *parv[])
{
#ifdef TRACK_VERSION
	struct Server *s = NULL;
	int i;

	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"[Server]                      [Version]");
	for (i = 0; i < S_TABLE_SIZE; i++)
	{
		for (s = serverlist[i]; s; s = s->next)
		{
			if (!IsSHide(s))
			{
				if (IsPubDelayed()) select_delay();
				privmsg(settings.pub_chan, StatServ.nick,
					MSG_PRIVMSG,
					"%-30s %s", s->name,
					(s->version && s->version[0]) ?
						s->version :
						"Unknown Version");
			}
		}
	}

	return 0;
#else
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"%s Disabled", PUB_VERSIONMAP);
	return 0;
#endif
}

static int p_tld(struct User *u, int parc, char *parv[])
{
	struct TLD *tmp;
	char *tld = parv[1];

	if (tld == NULL)
	{
		privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
			"Syntax: %s <tld>", PUB_TLD);
		return 0;
	}

        if (*tld == '*')
                tld++;
        if (*tld == '.')
                tld++;

	/* Catch it when it's just * or . */
	if (*tld == '\0')
	{
		privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
			"Syntax: %s <tld>", PUB_TLD);
		return 0;
	}

	tmp = findtld(tld);

	if (tmp == NULL)
	{
		privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
			"Top Level Domain %s does not exist.", tld);
		return 0;
	}

	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"%s (%s)", tmp->country, tmp->tld);
	if (tmp->native != NULL)
		privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
			"Native Name: %s", tmp->native);

	return 0;
}

static int p_tld_map(struct User *u, int parc, char *parv[])
{
	struct TLD *t;
	int i;

	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   TLD Number / Percent  Country");
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   --- ------   -------  -------");

	for (i = 0; i < T_TABLE_SIZE; i++)
	{
		for(t = tldlist[i]; t; t = t->next)
		{
			if (t->users != 0)
			{
				if (IsPubDelayed()) select_delay();
				privmsg(settings.pub_chan, StatServ.nick,
					MSG_PRIVMSG,
					"%9s %3d (%2.0f%%) -> %s", t->tld,
					t->users, (float)t->users /
					(float)netinfo.users * 100,
					t->country);
			}
		}
	}

	return 0;
}

static int p_url(struct User *u, int parc, char *parv[])
{
	if (!IsHtml())
	{
		privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
			"HTML Dump Disabled");
		return 0;
	}

	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"HTML Status URL: %s", settings.url);
	return 0;
}

static int p_help(struct User *u, int parc, char *parv[])
{
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"Supported commands: %s %s %s %s %s %s %s %s %s",
		PUB_MAP, PUB_VERSIONMAP, PUB_TLD, PUB_TLDMAP,
		PUB_COUNTRY, PUB_DAILY, PUB_NETSTATS,
		PUB_URL, PUB_HELP);
	return 0;
}

static int p_country(struct User *u, int parc, char *parv[])
{
	int i;
	struct TLD *t;
	char search[LINE];
	int matches = 0;

	if (parv[1] == NULL)
	{
		privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
			"Syntax: %s <countryname>", parv[0]);
		return 0;
	}

	/* Wrap the entered search string with *'s */
	snprintf(search, LINE, "*%s*", parv[1]);

	for (i = 0; i < T_TABLE_SIZE; i++)
	{
		for(t = tldlist[i]; t; t = t->next)
		{
			if (matches == COUNTRY_LIMIT)
			{
				privmsg(settings.pub_chan, StatServ.nick,
					MSG_PRIVMSG,
					"Too many matches, only %d returned",
					matches);
				return 0;
			}
			if (match(search, t->country) == 0)
			{
				matches++;
				privmsg(settings.pub_chan, StatServ.nick,
					MSG_PRIVMSG,
					"Match %d:  %s (%s)", matches,
					t->country, t->tld);
				if (IsPubDelayed()) select_delay();
				if (t->native != NULL)
				{
					privmsg(settings.pub_chan,StatServ.nick,
					    MSG_PRIVMSG,
					    "Native Name: %s", t->native);
					if (IsPubDelayed()) select_delay();
				}
				/* There can be multiple matches for the
				** same strings, consider Korea or Congo
				*/
			}
		}
	}

	return 0;
}

static int p_daily(struct User *u, int parc, char *parv[])
{
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"  Maximum Servers: %-2d %s", daily.servers,
		sftime(daily.t_servers));
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"  Maximum Users: %-2d %s",
		daily.users, sftime(daily.t_users));
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"  Maximum Opers: %-2d %s", daily.opers,
		sftime(daily.t_opers));
#ifdef IRCD_2_10
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"  Maximum Services: %-2d %s", daily.services,
		sftime(daily.t_services));
#endif
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"  Maximum Channels: %-2d %s", daily.channels,
		sftime(daily.t_channels));
	return 0;
}

static int p_netstats(struct User *u, int parc, char *parv[])
{
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Current Users: %ld", netinfo.users);
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Maximum Users: %-2ld %s", netinfo.maxusers,
		sftime(netinfo.t_maxusers));
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Current Opers: %d", netinfo.opers);
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Maximum Opers: %-2d %s", netinfo.maxopers,
		sftime(netinfo.t_maxopers));
	if (IsPubDelayed()) select_delay();
#if defined(IRCD_HYB7) || defined(IRCD_2_10) || defined(IRCD_BAHAMUT) || \
    defined(IRCD_HYBSVS)
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Users Set Away: %-2d", netinfo.away);
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
                "   Maximum Away: %-2d %s", netinfo.maxaway,
                sftime(netinfo.t_maxaway));
#endif
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Current Servers: %d", netinfo.servers);
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Maximum Servers: %-2d %s", netinfo.maxservers,
		sftime(netinfo.t_maxservers));
#ifdef IRCD_2_10
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Current Services: %d", netinfo.services);
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Maximum Services: %-2d %s",
		netinfo.maxservices, sftime(netinfo.t_maxservices));
#endif
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Current Channels: %d", netinfo.channels);
	if (IsPubDelayed()) select_delay();
	privmsg(settings.pub_chan, StatServ.nick, MSG_PRIVMSG,
		"   Maximum Channels: %-2d %s",
		netinfo.maxchannels, sftime(netinfo.t_maxchannels));
	return 0;
}

#endif
